use crate::*;

pub fn cell_diff_8_horizontal(prev_frame: &Frame, frame: &Frame) -> EncodedFrame {
	let cells_x = WIDTH / 8;
	let cells_y = HEIGHT / 8;
	let loss = 0;

	let mut bitmap: u32 = 0;
	let mut changed_pixels = Vec::new();

	for celly in 0..cells_y {
		for cellx in 0..cells_x {
			let mut changed = 0;
			let mut cell_contents = Vec::new();
			// todo try encoding the cells in frame space instead of one cell at a time
			for dx in 0..8 {
				for dy in 0..8 {
					let x = cellx * 8 + dx;
					let y = celly * 8 + dy;
					let pixel = frame[x][y];
					cell_contents.push(pixel);
					if pixel != prev_frame[x][y] {
						changed += 1;
					}
				}
			}
			if changed > loss {
				bitmap |= 1 << (cellx + cells_x * celly);
				changed_pixels.extend_from_slice(&cell_contents);
			}
		}
	}
	let mut data = vec![(bitmap >> 16) as u8, (bitmap >> 8) as u8, bitmap as u8];
	data.extend_from_slice(&rle_255_encode(&changed_pixels));
	EncodedFrame {
		encoding: Encoding::CellDiff8H,
		data,
	}
}

pub fn bg_strips_horizontal(_prev_frame: &Frame, frame: &Frame) -> EncodedFrame {
	let bg = most_common_pixel(frame);
	fn pack_strip(x: usize, y: usize, width: usize) -> [u8; 2] {
		// Y is 0..31 so will only need 5 bits
		// x is 0..42 so needs 6 bits
		// 5 bits remain for width
		// y: 1111100000000000
		let mut y_x_w: u16 = (y as u16) << 11;
		// x: 0000011111100000
		y_x_w |= (x as u16) << 5;
		// w: 0000000000011111
		y_x_w |= width as u16;
		[(y_x_w >> 8) as u8, (y_x_w as u8)]
	}

	let mut strips = Vec::new();
	'outer: for y in 0..HEIGHT {
		let mut strip_start = 0;
		let mut in_strip = false;
		for x in 0..WIDTH {
			let pixel = frame[x][y];
			if !in_strip && pixel != bg {
				in_strip = true;
				strip_start = x;
			}
			if in_strip {
				if pixel == bg {
					strips.push((strip_start, y, x - strip_start));
					in_strip = false;
				} else if x - strip_start == 31 {
					strips.push((strip_start, y, x - strip_start));
					strip_start = x;
				}
				if strips.len() == 127 {
					break 'outer;
				}
			}
		}
		if in_strip {
			strips.push((strip_start, y, WIDTH - strip_start));
		}
	}
	let mut frame_bytes = Vec::with_capacity(1 + strips.len() * 2);
	frame_bytes.push(bg << 7 | (strips.len() as u8));
	for (x, y, width) in strips {
		frame_bytes.extend_from_slice(&pack_strip(x, y, width));
	}
	EncodedFrame {
		encoding: Encoding::BGStripsH,
		data: frame_bytes,
	}
}

pub fn rle_diff_horizontal(prev_frame: &Frame, frame: &Frame) -> EncodedFrame {
	let mut pixels = Vec::new();
	for y in 0..HEIGHT {
		for x in 0..WIDTH {
			pixels.push(frame[x][y] ^ prev_frame[x][y]);
		}
	}
	EncodedFrame {
		encoding: Encoding::RLEDiffHorizontal,
		data: rle_255_encode(&pixels),
	}
}

pub fn rle_diff_vertical(prev_frame: &Frame, frame: &Frame) -> EncodedFrame {
	let mut pixels = Vec::new();
	for x in 0..WIDTH {
		for y in 0..HEIGHT {
			pixels.push(frame[x][y] ^ prev_frame[x][y]);
		}
	}
	EncodedFrame {
		encoding: Encoding::RLEDiffVertical,
		data: rle_255_encode(&pixels),
	}
}

pub fn rle_horizontal(_prev_frame: &Frame, frame: &Frame) -> EncodedFrame {
	let mut pixels = Vec::new();
	for y in 0..HEIGHT {
		for x in 0..WIDTH {
			pixels.push(frame[x][y]);
		}
	}
	EncodedFrame {
		encoding: Encoding::RLEHorizontal,
		data: rle_255_encode(&pixels),
	}
}

pub fn rle_vertical(_prev_frame: &Frame, frame: &Frame) -> EncodedFrame {
	let mut pixels = Vec::new();
	for x in 0..WIDTH {
		for y in 0..HEIGHT {
			pixels.push(frame[x][y]);
		}
	}
	EncodedFrame {
		encoding: Encoding::RLEVertical,
		data: rle_255_encode(&pixels),
	}
}

pub fn fill_white(_prev_frame: &Frame, _frame: &Frame) -> EncodedFrame {
	EncodedFrame {
		encoding: Encoding::FillWhite,
		data: Vec::new(),
	}
}

pub fn fill_black(_prev_frame: &Frame, _frame: &Frame) -> EncodedFrame {
	EncodedFrame {
		encoding: Encoding::FillBlack,
		data: Vec::new(),
	}
}
